""" Errors raised by service classes
"""

import json


class ServiceError(Exception):
    """ Errors raised from services
    """
    def __init__(self, status, response_text):
        self.status = status
        self.response_text = response_text

    def json_body(self):
        """ Parse the response body

            Returns the json dict or an empty dict if it fails
        """
        try:
            return json.loads(self.response_text)
        except json.decoder.JSONDecodeError:
            return {}


class InvalidDataError(ServiceError):   # pragma: no coverage
    """ Error when the service returns a 400 Bad request received
    """


class AuthenticationError(ServiceError):    # pragma: no coverage
    """ Error when the service returns a 401 Unauthorized
    """


class PaymentRequiredError(ServiceError):  # pragma: no coverage
    """ Error when the service returns a 402 Forbidden
    """


class AuthorizationError(ServiceError):  # pragma: no coverage
    """ Error when the service returns a 403 Forbidden
    """


class NotFoundError(ServiceError):  # pragma: no coverage
    """ Error when the service returns a 404 Not found
    """


class InternalServerError(ServiceError):  # pragma: no coverage
    """ Error when the service returns a 500 Not found
    """


class UnexpectedError(ServiceError):    # pragma: no coverage
    """ Error raised when the service returns an unexpected status code
    """



""" Tests for tracing
"""

from unittest.mock import MagicMock

from styler_rest_framework.tracing import config_tracer


def test_middleware():
    app = MagicMock()

    config_tracer(app, 'MyProject')

    app.middlewares.append.assert_called_once()



""" Context manager for a Session scope using Sql Alchemy.

    All functions that reads/writes data to a session in sql alchemy
should use this scope. It handles the initialization and cleanup of a
sql alchemy session within the given engine (connection).
"""

from contextlib import contextmanager

from sqlalchemy import orm


@contextmanager
def create(engine):
    """ Creates a context with a sql alchemy session
    """
    session_maker = orm.sessionmaker(bind=engine)
    session = session_maker()
    try:
        yield session
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()



""" Tests for session scope
"""

from unittest.mock import Mock, patch

from styler_rest_framework.datasource import session_scope
import pytest


class MySession:
    """ Represents a mock session
    """
    rollback = Mock()
    close = Mock()


@patch('sqlalchemy.orm.sessionmaker', Mock(return_value=MySession))
def test_session_in_context():
    with pytest.raises(ValueError):
        with session_scope.create(Mock()) as session:
            some_op = Mock(side_effect=ValueError('something wrong'))
            some_op()
    session.rollback.assert_called_once()
    session.close.assert_called_once()



""" Tests for service error
"""

import json

from styler_rest_framework.exceptions.services import ServiceError


class TestJsonBody:
    """ Tests for method json_body
    """
    def test_json_body(self):
        text = json.dumps({'key': 'error'})
        exception = ServiceError(status=400, response_text=text)

        result = exception.json_body()

        assert result == {'key': 'error'}

    def test_non_json_body(self):
        text = 'some text'
        exception = ServiceError(status=400, response_text=text)

        result = exception.json_body()

        assert result == {}




import logging

from aiohttp import web
from opencensus.common.transports.async_ import AsyncTransport
from opencensus.ext.stackdriver import trace_exporter as stackdriver_exporter
from opencensus.trace import attributes_helper
from opencensus.trace.propagation import google_cloud_format
import opencensus.trace.tracer


HTTP_HOST = attributes_helper.COMMON_ATTRIBUTES['HTTP_HOST']
HTTP_METHOD = attributes_helper.COMMON_ATTRIBUTES['HTTP_METHOD']
HTTP_PATH = attributes_helper.COMMON_ATTRIBUTES['HTTP_PATH']
HTTP_ROUTE = attributes_helper.COMMON_ATTRIBUTES['HTTP_ROUTE']
HTTP_URL = attributes_helper.COMMON_ATTRIBUTES['HTTP_URL']
HTTP_STATUS_CODE = attributes_helper.COMMON_ATTRIBUTES['HTTP_STATUS_CODE']


def initialize_tracer(project_id, context, propagator):
    exporter = stackdriver_exporter.StackdriverExporter(
        project_id=project_id,
        transport=AsyncTransport
    )
    tracer = opencensus.trace.tracer.Tracer(
        span_context=context,
        exporter=exporter,
        propagator=propagator,
        sampler=opencensus.trace.tracer.samplers.AlwaysOnSampler()
    )

    return tracer


def trace_middleware(project_id):
    @web.middleware
    async def middleware(request, handler):
        if handler.__name__ == 'health_check':
            return await handler(request)
        span = None
        tracer = None
        try:
            propagator = google_cloud_format.GoogleCloudFormatPropagator()
            span_context = propagator.from_headers(request.headers)
            tracer = initialize_tracer(project_id, span_context, propagator)
            span = tracer.start_span()
            span.name = handler.__name__
            tracer.add_attribute_to_current_span(
                HTTP_HOST, request.host
            )
            tracer.add_attribute_to_current_span(
                HTTP_METHOD, request.method
            )
            tracer.add_attribute_to_current_span(
                HTTP_PATH, request.path
            )
            tracer.add_attribute_to_current_span(
                HTTP_URL, str(request.url)
            )
            request['trace_header'] = propagator.to_headers(span_context)
        except:  # NOQA
            logging.exception('Could not initialize the tracer')

        try:
            response = await handler(request)
            if tracer:
                tracer.add_attribute_to_current_span(
                    HTTP_STATUS_CODE, response.status
                )
            return response
        finally:
            if tracer:
                tracer.end_span()
    return middleware


def config_tracer(app, project_id):
    app.middlewares.append(trace_middleware(project_id))



""" Tests for validation module
"""

from styler_rest_framework.validation import I18nValidatorMixin
from styler_rest_framework.validation import validators as va


class MyModel(I18nValidatorMixin):
    validates = [
        ('name', va.is_required()),
        ('age', va.is_required(), va.is_less_than_number(100))
    ]


def test_validation():
    model = MyModel()
    model.name = None
    model.age = 200

    result, errors = model.is_valid()

    assert not result
    assert errors == {
        'mymodel.age': '100より小さい値にしてください',
        'mymodel.name': '入力してください'
    }


def test_validation_i18n():
    model = MyModel()
    model.name = None
    model.age = 200

    result, errors = model.is_valid(locale='en')

    assert not result
    assert errors == {
        'mymodel.age': 'Should be less than 100',
        'mymodel.name': 'Required field'
    }


def test_valid():
    model = MyModel()
    model.name = 'John'
    model.age = 67

    result, errors = model.is_valid()

    assert result
    assert errors == {}



""" Handles i18n
"""

import os

from i18n import t
import i18n


dir_path = os.path.dirname(os.path.realpath(__file__))
i18n.set('enable_memoization', True)
i18n.load_path.append(f'{dir_path}/labels')
i18n.set('locale', 'ja')
i18n.set('fallback', 'ja')


def get(key, locale='ja', **kwargs):  # pragma: no coverage
    return t(key, locale=locale, **kwargs)


def load_path(path):  # pragma: no coverage
    i18n.load_path.append(path)



""" Request scope handles the identification of the user and request options
"""

from styler_identity import Identity


class RequestScope(Identity):
    """ Carries information about the request scope
    """
    def __init__(self, request):
        self.request = request
        token = token = get_token_auth_header(request)
        super().__init__(token)

    def localization(self):
        return self.request.headers.get('Accept-Language', 'ja')

    def trace_header(self):
        return self.request.headers.get('trace_header', {})


def get_token_auth_header(request):
    """ Obtains the Access Token from the Authorization Header
    """
    auth_header = request.headers.get('Authorization')
    if not auth_header:
        raise ValueError('Authorization header is expected')

    parts = auth_header.split()

    if parts[0].lower() != 'bearer':
        raise ValueError('Authorization header must start with Bearer')
    elif len(parts) == 1:
        raise ValueError('Token not found')
    elif len(parts) > 2:
        raise ValueError('Authorization header must be Bearer token')

    token = parts[1]
    return token



""" Tests for message module
"""

from styler_rest_framework import message


def test_18_n():
    assert message.get(
        'validation.required_value', 'en') == 'Required field'
    assert message.get(
        'validation.required_value', 'ja') == '入力してください'
    assert message.get(
        'validation.required_value') == '入力してください'
    assert message.get(
        'validation.required_value', 'es') == '入力してください'



""" Paginator
"""

from math import ceil


class InvalidParameterError(ValueError):
    """ Raised when page or limit is invalid
    """


class Paginator:
    """ Handles pagination
    """

    def __init__(self, limit=20, page=1):
        try:
            if not page:
                page = 1
            self._current = int(page)
        except ValueError:
            raise InvalidParameterError('Invalid page')

        try:
            if not limit:
                limit = 20
            self._limit = int(limit)
        except ValueError:
            raise InvalidParameterError('Invalid limit')

        if self._current < 1:
            raise InvalidParameterError('Invalid page')
        if self._limit < 1:
            raise InvalidParameterError('Invalid limit')

        self._count = None
        self._total_pages = None

    def get_page(self, query):
        """ Retrieve a page of items

            Args:
                query: query object
            Returns:
                A list of items within the boundaries set by the pagination
        """
        self._count = query.count()
        self._total_pages = int(ceil(self._count / float(self._limit)))
        offset = (self._current - 1) * self._limit
        if self._total_pages > 0 and self._current > self._total_pages:
            raise InvalidParameterError('Invalid page')
        return query.offset(offset).limit(self._limit).all()

    def get_info(self):
        """ Retrieves information about the pagination status
        """
        nex_page = (
            self._current + 1
            if self._current < self._total_pages
            else None
        )
        return {
            'total_pages': self._total_pages,
            'total_number_of_items': self._count,
            'current_page': self._current,
            'previous_page': self._current - 1 if self._current > 1 else None,
            'next_page': nex_page
        }



""" Module for validations (i18n enabled)
"""

from styler_rest_framework import message
from styler_validation import ValidatorMixin
from styler_validation import validators  # NOQA


class I18nValidatorMixin(ValidatorMixin):
    """ Translates validation messages
    """
    def is_valid(self, locale='ja', **kwargs):
        result, errors = super().is_valid(**kwargs)
        if result:
            return result, errors
        localized_errors = {}
        for key, msg in errors.items():
            if len(msg) == 2:
                localized_errors[key] = message.get(
                    f'validation.{msg[0]}', locale=locale, value=msg[1])
            else:
                localized_errors[key] = message.get(
                    f'validation.{msg[0]}', locale=locale)
        return result, localized_errors



""" Module for business classes
"""

from functools import wraps

from styler_rest_framework.datasource import session_scope


def session_aware(func):
    """ Decorator to wrap the method inside a session aware context.

        It also injects the db_session in the function, necessary to
        control flush, commit, and rollbacks.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        with session_scope.create(args[0].db_engine) as db_session:
            try:
                kwargs['db_session'] = db_session
                return func(*args, **kwargs)
            except:  # NOQA
                db_session.close()
                raise
    return wrapper


class BaseBusiness:
    """ Base class for business classes
    """



""" Tests for HTTP handler
"""

from unittest.mock import Mock
import json

from styler_rest_framework.services import HTTPHandler
from styler_rest_framework.exceptions.services import (
    AuthenticationError,
    AuthorizationError,
    InternalServerError,
    InvalidDataError,
    NotFoundError,
    PaymentRequiredError,
    UnexpectedError,
)
import pytest


class IdentityMock:
    """ Mock an identity instance
    """
    def token(self):
        return 'aaa.aaaaa.aaa'

    def trace_header(self):
        return {}

    def localization(self):
        return 'ja'


class TestInit:
    """ Tests for init
    """
    def test_with_identity(self):
        iden = IdentityMock()
        handler = HTTPHandler(identity=iden)

        assert 'Authorization' in handler.headers
        assert handler.headers['Authorization'] == f'Bearer {iden.token()}'

    def test_with_custom_headers(self):
        handler = HTTPHandler(
            identity=IdentityMock(),
            headers={'my-custom-header': '1234'}
        )

        assert 'my-custom-header' in handler.headers
        assert handler.headers['my-custom-header'] == '1234'


class TestPost:
    """ Test method post
    """
    async def test_post(self, aresponses):
        aresponses.add(
            'some.url',
            '/resources',
            'POST',
            aresponses.Response(
                status=201,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())

        result = await handler.post(
            'https://some.url/resources',
            {'name': 'something'}
        )

        assert result == {'id': '1234'}

    test_cases = [
        (400, InvalidDataError),
        (401, AuthenticationError),
        (402, PaymentRequiredError),
        (403, AuthorizationError),
        (404, NotFoundError),
        (500, InternalServerError),
        (986, UnexpectedError),
    ]

    @pytest.mark.parametrize("status, exception", test_cases)
    async def test_error_responses(self, status, exception, aresponses):
        aresponses.add(
            'some.url',
            '/resources',
            'POST',
            aresponses.Response(
                status=status,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())
        with pytest.raises(exception):
            _ = await handler.post(
                'https://some.url/resources',
                {'name': 'something'}
            )

    async def test_custom_handler(self, aresponses):
        aresponses.add(
            'some.url',
            '/resources',
            'POST',
            aresponses.Response(
                status=402,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())

        with pytest.raises(ValueError):
            _ = await handler.post(
                'https://some.url/resources',
                {'name': 'something'},
                error_handlers={402: Mock(side_effect=ValueError())}
            )


class TestPut:
    """ Test method put
    """
    async def test_put(self, aresponses):
        aresponses.add(
            'some.url',
            '/resources/1234',
            'PUT',
            aresponses.Response(
                status=200,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())

        result = await handler.put(
            'https://some.url/resources/1234',
            {'name': 'something'}
        )

        assert result == {'id': '1234'}

    test_cases = [
        (400, InvalidDataError),
        (401, AuthenticationError),
        (402, PaymentRequiredError),
        (403, AuthorizationError),
        (404, NotFoundError),
        (500, InternalServerError),
        (986, UnexpectedError),
    ]

    @pytest.mark.parametrize("status, exception", test_cases)
    async def test_error_responses(self, status, exception, aresponses):
        aresponses.add(
            'some.url',
            '/resources/1234',
            'PUT',
            aresponses.Response(
                status=status,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())
        with pytest.raises(exception):
            _ = await handler.put(
                'https://some.url/resources/1234',
                {'name': 'something'}
            )

    async def test_custom_handler(self, aresponses):
        aresponses.add(
            'some.url',
            '/resources/1234',
            'PUT',
            aresponses.Response(
                status=402,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())

        with pytest.raises(ValueError):
            _ = await handler.put(
                'https://some.url/resources/1234',
                {'name': 'something'},
                error_handlers={402: Mock(side_effect=ValueError())}
            )


class TestPatch:
    """ Test method patch
    """
    async def test_patch(self, aresponses):
        aresponses.add(
            'some.url',
            '/resources/1234',
            'PATCH',
            aresponses.Response(
                status=200,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())

        result = await handler.patch(
            'https://some.url/resources/1234',
            {'name': 'something'}
        )

        assert result == {'id': '1234'}

    test_cases = [
        (400, InvalidDataError),
        (401, AuthenticationError),
        (402, PaymentRequiredError),
        (403, AuthorizationError),
        (404, NotFoundError),
        (500, InternalServerError),
        (986, UnexpectedError),
    ]

    @pytest.mark.parametrize("status, exception", test_cases)
    async def test_error_responses(self, status, exception, aresponses):
        aresponses.add(
            'some.url',
            '/resources',
            'PATCH',
            aresponses.Response(
                status=status,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())
        with pytest.raises(exception):
            _ = await handler.patch(
                'https://some.url/resources',
                {'name': 'something'}
            )

    async def test_custom_handler(self, aresponses):
        aresponses.add(
            'some.url',
            '/resources',
            'PATCH',
            aresponses.Response(
                status=402,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())

        with pytest.raises(ValueError):
            _ = await handler.patch(
                'https://some.url/resources',
                {'name': 'something'},
                error_handlers={402: Mock(side_effect=ValueError())}
            )


class TestGet:
    """ Test method get
    """
    async def test_get(self, aresponses):
        aresponses.add(
            'some.url',
            '/resources',
            'GET',
            aresponses.Response(
                status=200,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())

        result = await handler.get(
            'https://some.url/resources'
        )

        assert result == {'id': '1234'}

    test_cases = [
        (400, InvalidDataError),
        (401, AuthenticationError),
        (402, PaymentRequiredError),
        (403, AuthorizationError),
        (404, NotFoundError),
        (500, InternalServerError),
        (986, UnexpectedError),
    ]

    @pytest.mark.parametrize("status, exception", test_cases)
    async def test_error_responses(self, status, exception, aresponses):
        aresponses.add(
            'some.url',
            '/resources',
            'GET',
            aresponses.Response(
                status=status,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())
        with pytest.raises(exception):
            _ = await handler.get(
                'https://some.url/resources'
            )

    async def test_custom_handler(self, aresponses):
        aresponses.add(
            'some.url',
            '/resources',
            'GET',
            aresponses.Response(
                status=402,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())

        with pytest.raises(ValueError):
            _ = await handler.get(
                'https://some.url/resources',
                error_handlers={402: Mock(side_effect=ValueError())}
            )


class TestDelete:
    """ Test method delete
    """
    async def test_delete(self, aresponses):
        aresponses.add(
            'some.url',
            '/resources',
            'DELETE',
            aresponses.Response(
                status=200,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())

        result = await handler.delete(
            'https://some.url/resources'
        )

        assert result == {'id': '1234'}

    test_cases = [
        (400, InvalidDataError),
        (401, AuthenticationError),
        (402, PaymentRequiredError),
        (403, AuthorizationError),
        (404, NotFoundError),
        (500, InternalServerError),
        (986, UnexpectedError),
    ]

    @pytest.mark.parametrize("status, exception", test_cases)
    async def test_error_responses(self, status, exception, aresponses):
        aresponses.add(
            'some.url',
            '/resources',
            'DELETE',
            aresponses.Response(
                status=status,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())
        with pytest.raises(exception):
            _ = await handler.delete(
                'https://some.url/resources'
            )

    async def test_custom_handler(self, aresponses):
        aresponses.add(
            'some.url',
            '/resources',
            'DELETE',
            aresponses.Response(
                status=402,
                text=json.dumps({'id': '1234'}),
                content_type='application/json'
            )
        )
        handler = HTTPHandler(IdentityMock())

        with pytest.raises(ValueError):
            _ = await handler.delete(
                'https://some.url/resources',
                error_handlers={402: Mock(side_effect=ValueError())}
            )



""" Configures logging with stackdriver
"""

import logging  # pragma: no coverage

import google.cloud.logging  # pragma: no coverage


def setup_logging(level=logging.INFO):  # pragma: no coverage
    """ Setup logging
    """
    client = google.cloud.logging.Client()
    excluded_logger_defaults = (
        "google.cloud",
        "google.auth",
        "google_auth_httplib2"
    )

    # don't propagate excluded loggers (and don't send them to stderr either)
    for logger_name in excluded_logger_defaults:
        logging.getLogger(logger_name).propagate = False
        logging.getLogger(logger_name).addHandler(logging.NullHandler())

    handler = client.get_default_handler()
    handler.setLevel(level)
    root = logging.getLogger()
    root.addHandler(handler)
    root.setLevel(level)



""" Tests for the Paginator
"""

import pytest
from unittest.mock import Mock, call

from styler_rest_framework.helpers.paginator import \
    Paginator, InvalidParameterError


class TestInit:
    """ Tests for constructor
    """
    def test_init_a_paginator(self):
        pag = Paginator(limit=10, page=1)

        assert isinstance(pag, Paginator)

    def test_init_a_paginator_with_none(self):
        pag = Paginator(limit=None, page=None)

        assert isinstance(pag, Paginator)
        assert pag._current == 1
        assert pag._limit == 20

    def test_invalid_limit(self):
        with pytest.raises(InvalidParameterError) as expected:
            Paginator(limit=-1, page=1)
        assert str(expected.value) == 'Invalid limit'

    def test_invalid_limit_type(self):
        with pytest.raises(InvalidParameterError) as expected:
            Paginator(limit='aaa', page=1)
        assert str(expected.value) == 'Invalid limit'

    def test_invalid_page(self):
        with pytest.raises(InvalidParameterError) as expected:
            Paginator(limit=10, page=-1)
        assert str(expected.value) == 'Invalid page'

    def test_invalid_page_type(self):
        with pytest.raises(InvalidParameterError) as expected:
            Paginator(limit=10, page='aaa')
        assert str(expected.value) == 'Invalid page'


class TestGetPage:
    """ Tests for get_page
    """
    def test_normal_flow(self):
        """ It should return the item page without errors
        """
        query = Mock()
        query.count.return_value = 100
        query.offset.return_value = query
        query.limit.return_value = query
        query.all.return_value = [1, 2, 3]
        pag = Paginator(limit=10, page=1)

        result = pag.get_page(query)

        assert result == [1, 2, 3]
        assert query.count.call_count == 1
        assert call(0) in query.offset.mock_calls
        assert call(10) in query.limit.mock_calls

    def test_invalid_page(self):
        """ It should raise an exception
        """
        query = Mock()
        query.count.return_value = 5
        query.offset.return_value = query
        query.limit.return_value = query
        query.all.return_value = [1, 2, 3]
        pag = Paginator(limit=10, page=2)

        with pytest.raises(InvalidParameterError) as expected:
            _ = pag.get_page(query)

        assert str(expected.value) == 'Invalid page'
        assert query.count.call_count == 1
        assert query.offset.call_count == 0
        assert query.limit.call_count == 0


class TestGetInfo:
    """ Tests for get_info
    """
    def test_get_info(self):
        query = Mock()
        query.count.return_value = 100
        query.offset.return_value = query
        query.limit.return_value = query
        query.all.return_value = [1, 2, 3]
        pag = Paginator(limit=10, page=3)

        pag.get_page(query)
        info = pag.get_info()

        assert info['total_pages'] == 10
        assert info['total_number_of_items'] == 100
        assert info['current_page'] == 3
        assert info['previous_page'] == 2
        assert info['next_page'] == 4

    def test_none_previous_page(self):
        """ Should return None in the previous page if current page == 1
        """
        query = Mock()
        query.count.return_value = 100
        query.offset.return_value = query
        query.limit.return_value = query
        query.all.return_value = [1, 2, 3]
        pag = Paginator(limit=10, page=1)

        pag.get_page(query)
        info = pag.get_info()

        assert info['total_pages'] == 10
        assert info['total_number_of_items'] == 100
        assert info['current_page'] == 1
        assert info['previous_page'] is None
        assert info['next_page'] == 2

    def test_none_next_page(self):
        """ Should return None in the next page if current page == total_pages
        """
        query = Mock()
        query.count.return_value = 100
        query.offset.return_value = query
        query.limit.return_value = query
        query.all.return_value = [1, 2, 3]
        pag = Paginator(limit=10, page=10)

        pag.get_page(query)
        info = pag.get_info()

        assert info['total_pages'] == 10
        assert info['total_number_of_items'] == 100
        assert info['current_page'] == 10
        assert info['previous_page'] == 9
        assert info['next_page'] is None



""" Tests for request_scope module
"""

from unittest.mock import Mock

from styler_rest_framework.controllers.request_scope import (
    RequestScope,
    get_token_auth_header,
)
import pytest


class TestGetTokenAuthHeader:
    """ Tests for get_token_auth_header
    """
    def test_get_token(self):
        """ Should return the token
        """
        request = Mock()
        request.headers.get.return_value = 'Bearer TOKEN'

        token = get_token_auth_header(request)

        assert token == 'TOKEN'

    def test_without_header(self):
        """ Should raise an exception if there is no Authorization header
        """
        request = Mock()
        request.headers.get.return_value = None

        with pytest.raises(ValueError) as expected:
            get_token_auth_header(request)
        assert str(expected.value) == 'Authorization header is expected'

    def test_without_bearer(self):
        """ Should raise an exception if there is no bearer
        """
        request = Mock()
        request.headers.get.return_value = 'token TOKEN'

        with pytest.raises(ValueError) as expected:
            get_token_auth_header(request)
        assert str(expected.value) == \
            'Authorization header must start with Bearer'

    def test_without_token(self):
        """ Should raise an exception if the auth is invalid
        """
        request = Mock()
        request.headers.get.return_value = 'Bearer'

        with pytest.raises(ValueError) as expected:
            get_token_auth_header(request)
        assert str(expected.value) == 'Token not found'

    def test_invalid_auth(self):
        """ Should raise an exception if the Authorization is invalid
        """
        request = Mock()
        request.headers.get.return_value = 'Bearer TOKEN TOKEN'

        with pytest.raises(ValueError) as expected:
            get_token_auth_header(request)
        assert str(expected.value) == \
            'Authorization header must be Bearer token'


class MockHeaders:
    def __init__(self, locale, header=None):
        self.locale = locale
        self.header = header

    def get(self, key, default=None):
        if key == 'Authorization':
            return (
                'Bearer '
                'eyJhbGciOiJSUzI1NiIsImtpZCI6Im'
                'MzZjI3NjU0MmJmZmU0NWU5OG'
                'MyMGQ2MDNlYmUyYmExMTc2ZWRhMzMiLCJ0e'
                'XAiOiJKV1QifQ.eyJzeXN0ZW1fYWRta'
                'W4iOmZhbHNlLCJjbGFpbXMiOnsic2hvcHMi'
                'OlsiN2YwNzZjNDgtMGU0Yy00MDQ0LTl'
                'kNDctMjM0MjM0MzI0MjM0IiwiN2YwNzZjND'
                'gtMGU0Yy00MDQ0LTlkNDctMzg4OTk5M'
                'Dk4MDkwIl0sIm9yZ2FuaXphdGlvbnMiOltd'
                'fSwiaXNzIjoiaHR0cHM6Ly9zZWN1cmV'
                '0b2tlbi5nb29nbGUuY29tL2ZhY3ktZGV2ZW'
                'xvcG1lbnQiLCJhdWQiOiJmYWN5LWRld'
                'mVsb3BtZW50IiwiYXV0aF90aW1lIjoxNTky'
                'ODg2Mjg5LCJ1c2VyX2lkIjoiNm5hZjg'
                'wanR5YlZSdVAyMmVZY0lZR1gyS0pLMiIsIn'
                'N1YiI6IjZuYWY4MGp0eWJWUnVQMjJlW'
                'WNJWUdYMktKSzIiLCJpYXQiOjE1OTI4ODYy'
                'ODksImV4cCI6MTU5Mjg4OTg4OSwiZW1'
                'haWwiOiJicnVuby5zdWdhbm9Ac3R5bGVyLm'
                'xpbmsiLCJlbWFpbF92ZXJpZmllZCI6Z'
                'mFsc2UsImZpcmViYXNlIjp7ImlkZW50aXRp'
                'ZXMiOnsiZW1haWwiOlsiYnJ1bm8uc3V'
                'nYW5vQHN0eWxlci5saW5rIl19LCJzaWduX2'
                'luX3Byb3ZpZGVyIjoiY3VzdG9tIn19.'
                'VNJtVQ9nmvbpUkMM95cdN06O3NDUqdBwEgm'
                'czfWKR2Vt5fOFQGtFykEovEmajdVHZO'
                'm4tbJlbfoILu0i_GsO-7jaoYNBOC9-tIBiD'
                '8pvUCeL4iQj2jhpvR6z3PEtTyA7i7Ae'
                'VUj9wuxrzYm0wtcLTR1vSEp4jTj3Nw5KD4a'
                'dPq4IF63SKqywNLkmgURuQnF-WlaZBn'
                'g2CK0AMddY0Qt-1Z0IFUp-i_tXkGMOPBjXs'
                'wq1kPxyZ2xtNWZjQZ5sWNw2dmkC2wyf'
                'EmnQ5Zm9UJvrS6tiCigTGkF43PhwTl26K9z'
                'twOtekrpOiwap-c55rUG8TbobpmGGkF'
                'i9Q3r5dHH-ag'
            )
        if key == 'Accept-Language':
            return self.locale or default
        if key == 'trace_header':
            return self.header or {}


class TestLocalization:
    """ Tests for localization
    """
    def test_localization(self):
        request = Mock()
        request.headers = MockHeaders('en')
        req_scope = RequestScope(request)

        localization = req_scope.localization()

        assert localization == 'en'

    def test_localization_none(self):
        request = Mock()
        request.headers = MockHeaders(None)
        req_scope = RequestScope(request)

        localization = req_scope.localization()

        assert localization == 'ja'


class TestTraceHeader:
    """ Tests for trace_header
    """
    def test_trace_header(self):
        request = Mock()
        request.headers = MockHeaders('en', {'aa': 'bb'})
        req_scope = RequestScope(request)

        trace_header = req_scope.trace_header()

        assert trace_header == {'aa': 'bb'}

    def test_trace_header_none(self):
        request = Mock()
        request.headers = MockHeaders(None)
        req_scope = RequestScope(request)

        trace_header = req_scope.trace_header()

        assert trace_header == {}



""" Tests for server defaults
"""

from unittest.mock import Mock

from styler_rest_framework.helpers import server_defaults


class TestAddMiddlewares:
    def test_add_middlewares(self):
        app = Mock()
        app.middlewares = []

        server_defaults.add_middlewares(app)

        assert len(app.middlewares) == 2
        assert callable(app.middlewares[0])
        assert callable(app.middlewares[1])



""" Module for controllers
"""

import json
import logging

from aiohttp import web
from styler_rest_framework.exceptions.services import (
    AuthenticationError,
    AuthorizationError,
    InternalServerError,
    InvalidDataError,
    NotFoundError,
    PaymentRequiredError,
    UnexpectedError,
)
from styler_rest_framework.exceptions.business import (
    InternalError,
    PermissionDeniedError,
    ResourceNotFoundError,
    ValidationError,
)
from styler_rest_framework.controllers.request_scope import RequestScope


class BaseController:
    """ Base class for controllers
    """

    def get_identity(self, request):
        """ Get the identity of the logged user
        """
        try:
            return RequestScope(request)
        except ValueError:
            self.bad_request({'jwt': 'invalid'})

    def resp_for_creation(self, new_id):
        """ Default response for creation methods
        """
        return web.json_response({'id': new_id}, status=201)

    def response_ok(self, data):
        """ Default response for 200 OK
        """
        return web.json_response(data)

    def resp_for_listing(self, paginator, items):
        """ Default response for listings with pagination
        """
        return web.json_response({
            'items': items,
            'pagination': paginator.get_info()
        })

    def payment_required(self, code, reason):
        """ Default response for Error code 402
        """
        error = {
            'code': code,
            'reason': reason
        }
        logging.error('Payment required error: %s', error)
        raise web.HTTPPaymentRequired(
            text=json.dumps(error),
            content_type='application/json')

    def bad_request(self, errors, code='validation_error'):
        """ Default response for Error code 400
        """
        error = {
            'code': code,
            'reason': errors
        }
        logging.error('Bad request error: %s', error)
        raise web.HTTPBadRequest(
            text=json.dumps(error),
            content_type='application/json')

    def not_found(self, msg=''):
        """ Default response for Error code 404
        """
        logging.error('Not found error error: %s', msg)
        raise web.HTTPNotFound(
            content_type='application/json')

    def unauthorized(self, msg=''):
        """ Default response for Error code 401
        """
        logging.error('Unauthorized error: %s', msg)
        raise web.HTTPUnauthorized(
            content_type='application/json')

    def forbidden(self, msg=''):
        """ Default response for Error code 403
        """
        logging.error('Forbidden error: %s', msg)
        raise web.HTTPForbidden(
            content_type='application/json')

    def internal_server_error(self, exception, msg=''):
        """ Default response for Error code 500
        """
        logging.exception('Internal Server error: %s', str(exception))
        raise web.HTTPInternalServerError(
            content_type='application/json')

    def handle_service_errors(self, ex):
        """ Default method for handling service related errors
        """
        if isinstance(ex, InvalidDataError):
            self.bad_request(ex.json_body().get('reason'))
        elif isinstance(ex, AuthenticationError):
            self.unauthorized()
        elif isinstance(ex, PaymentRequiredError):
            body = ex.json_body()
            self.payment_required(body.get('code'), body.get('reason'))
        elif isinstance(ex, AuthorizationError):
            self.forbidden()
        elif isinstance(ex, NotFoundError):
            self.not_found()
        elif isinstance(ex, InternalServerError):
            self.internal_server_error(exception=ex)
        elif isinstance(ex, UnexpectedError):
            self.internal_server_error(exception=ex)
        else:
            raise ex

    def handle_business_errors(self, ex):
        """ Default method for handling business related errors
        """
        if isinstance(ex, ValidationError):
            self.bad_request(ex.errors)
        elif isinstance(ex, PermissionDeniedError):
            self.forbidden()
        elif isinstance(ex, ResourceNotFoundError):
            self.not_found()
        elif isinstance(ex, InternalError):
            self.internal_server_error(exception=ex)
        else:
            raise ex



""" Tests for the base controller
"""

from unittest.mock import MagicMock, Mock, patch
import json

from aiohttp import web
from styler_rest_framework.controllers import BaseController
from styler_rest_framework.controllers.request_scope import RequestScope
from styler_rest_framework.exceptions.services import (
    AuthenticationError,
    AuthorizationError,
    InternalServerError,
    InvalidDataError,
    NotFoundError,
    PaymentRequiredError,
    UnexpectedError,
)
from styler_rest_framework.exceptions.business import (
    InternalError,
    PermissionDeniedError,
    ResourceNotFoundError,
    ValidationError,
)
import pytest


class TestResponses:
    """ Tests for responses
    """
    @patch('aiohttp.web.json_response')
    def test_resp_for_creation(self, mocked_resp):
        """ HTTP 201 with new id
        """
        base = BaseController()
        mocked_resp.return_value = 'abc'

        resp = base.resp_for_creation('abc')

        assert resp == 'abc'
        mocked_resp.assert_called_with({'id': 'abc'}, status=201)

    @patch('aiohttp.web.json_response')
    def test_response_ok(self, mocked_resp):
        """ Default HTTP 200 with data
        """
        base = BaseController()
        mocked_resp.return_value = 'abc'

        resp = base.response_ok('abc')

        assert resp == 'abc'
        mocked_resp.assert_called_with('abc')

    @patch('aiohttp.web.json_response')
    def test_resp_for_listing(self, mocked_resp):
        """ Default HTTP 200 with list
        """
        base = BaseController()
        paginator = Mock()
        paginator.get_info.return_value = 'aaa'
        mocked_resp.return_value = 'abc'

        resp = base.resp_for_listing(paginator, ['1234'])

        assert resp == 'abc'
        mocked_resp.assert_called_with({
            'items': ['1234'],
            'pagination': 'aaa'
        })

    @patch('logging.error')
    def test_bad_request(self, mocked_logger):
        """ Default HTTP 400 with errors
        """
        base = BaseController()
        errors = {
            'error': 'general error'
        }

        with pytest.raises(web.HTTPBadRequest) as expected:
            base.bad_request(errors)
        mocked_logger.assert_called_once()
        resp = json.loads(expected.value.text)
        assert resp['reason'] == errors

    @patch('logging.error')
    def test_not_found(self, mocked_logger):
        """ Default HTTP 404 with errors
        """
        base = BaseController()

        with pytest.raises(web.HTTPNotFound) as expected:
            base.not_found()
        mocked_logger.assert_called_once()
        assert expected.value.status == 404

    @patch('logging.error')
    def test_forbidden(self, mocked_logger):
        """ Default HTTP 403 with errors
        """
        base = BaseController()

        with pytest.raises(web.HTTPForbidden) as expected:
            base.forbidden()
        mocked_logger.assert_called_once()
        assert expected.value.status == 403

    @patch('logging.error')
    def test_unauthorized(self, mocked_logger):
        """ Default HTTP 401 with errors
        """
        base = BaseController()

        with pytest.raises(web.HTTPUnauthorized) as expected:
            base.unauthorized()
        mocked_logger.assert_called_once()
        assert expected.value.status == 401

    @patch('logging.error')
    def test_payment_required(self, mocked_logger):
        """ Default HTTP 402 with errors
        """
        base = BaseController()

        with pytest.raises(web.HTTPPaymentRequired) as expected:
            base.payment_required(code='aaa', reason='bbb')
        mocked_logger.assert_called_once()
        assert expected.value.status == 402

    @patch('logging.exception')
    def test_internal_server_error(self, mocked_logger):
        """ Default HTTP 500 with errors
        """
        base = BaseController()

        with pytest.raises(web.HTTPInternalServerError) as expected:
            base.internal_server_error(ValueError())
        mocked_logger.assert_called_once()
        assert expected.value.status == 500


class TestHandleServiceErrors:
    """ Tests for handle_service_errors
    """
    async def test_invalid_data(self):
        base = BaseController()
        base.bad_request = MagicMock('aaa')
        error = {
            'code': 'some_code',
            'reason': {'error': 'something'}
        }

        base.handle_service_errors(
            InvalidDataError(400, json.dumps(error)))

        base.bad_request.assert_called_with({'error': 'something'})

    async def test_authentication_error(self):
        base = BaseController()
        base.unauthorized = MagicMock('aaa')

        base.handle_service_errors(AuthenticationError(401, ''))

        base.unauthorized.assert_called_with()

    async def test_payment_required_error(self):
        base = BaseController()
        base.payment_required = MagicMock('aaa')
        error = {
            'code': 'some_code',
            'reason': {'error': 'something'}
        }

        base.handle_service_errors(
            PaymentRequiredError(402, json.dumps(error)))

        base.payment_required.assert_called_with(
            'some_code', {'error': 'something'})

    async def test_authorization_error(self):
        base = BaseController()
        base.forbidden = MagicMock('aaa')

        base.handle_service_errors(AuthorizationError(403, ''))

        base.forbidden.assert_called_with()

    async def test_not_found_error(self):
        base = BaseController()
        base.not_found = MagicMock('aaa')

        base.handle_service_errors(NotFoundError(404, ''))

        base.not_found.assert_called_with()

    async def test_internal_server_error(self):
        base = BaseController()
        base.internal_server_error = MagicMock()
        exception = InternalServerError(500, '')

        base.handle_service_errors(exception)

        base.internal_server_error.assert_called_with(exception=exception)

    async def test_unexpected_error(self):
        base = BaseController()
        base.internal_server_error = MagicMock()
        exception = UnexpectedError(500, '')

        base.handle_service_errors(exception)

        base.internal_server_error.assert_called_with(exception=exception)

    async def test_other_exceptions_error(self):
        base = BaseController()

        with pytest.raises(ValueError):
            base.handle_service_errors(ValueError())


class TestHandleBusinessErrors:
    """ Tests for handle_business_errors
    """
    def test_invalid_data(self):
        base = BaseController()
        base.bad_request = MagicMock('aaa')
        error = {'field': 'invalid'}

        base.handle_business_errors(ValidationError(error))

        base.bad_request.assert_called_with({'field': 'invalid'})

    def test_forbidden_error(self):
        base = BaseController()
        base.forbidden = MagicMock('aaa')

        base.handle_business_errors(PermissionDeniedError())

        base.forbidden.assert_called_with()

    def test_not_found_error(self):
        base = BaseController()
        base.not_found = MagicMock('aaa')

        base.handle_business_errors(ResourceNotFoundError())

        base.not_found.assert_called_with()

    def test_internal_error(self):
        base = BaseController()
        base.internal_server_error = MagicMock('aaa')
        exception = InternalError()

        base.handle_business_errors(exception)

        base.internal_server_error.assert_called_with(exception=exception)

    def test_other_exceptions_error(self):
        base = BaseController()

        with pytest.raises(ValueError):
            base.handle_business_errors(ValueError())


class TestGetIdentity:
    """ Tests for BaseController.get_identity
    """
    custom_token = (
        'eyJhbGciOiJSUzI1NiIsImtpZCI6ImMzZjI3NjU0MmJmZmU0NWU5OG'
        'MyMGQ2MDNlYmUyYmExMTc2ZWRhMzMiLCJ0eXAiOiJKV1QifQ.eyJzeXN0ZW1fYWRta'
        'W4iOmZhbHNlLCJjbGFpbXMiOnsic2hvcHMiOlsiN2YwNzZjNDgtMGU0Yy00MDQ0LTl'
        'kNDctMjM0MjM0MzI0MjM0IiwiN2YwNzZjNDgtMGU0Yy00MDQ0LTlkNDctMzg4OTk5M'
        'Dk4MDkwIl0sIm9yZ2FuaXphdGlvbnMiOltdfSwiaXNzIjoiaHR0cHM6Ly9zZWN1cmV'
        '0b2tlbi5nb29nbGUuY29tL2ZhY3ktZGV2ZWxvcG1lbnQiLCJhdWQiOiJmYWN5LWRld'
        'mVsb3BtZW50IiwiYXV0aF90aW1lIjoxNTkyODg2Mjg5LCJ1c2VyX2lkIjoiNm5hZjg'
        'wanR5YlZSdVAyMmVZY0lZR1gyS0pLMiIsInN1YiI6IjZuYWY4MGp0eWJWUnVQMjJlW'
        'WNJWUdYMktKSzIiLCJpYXQiOjE1OTI4ODYyODksImV4cCI6MTU5Mjg4OTg4OSwiZW1'
        'haWwiOiJicnVuby5zdWdhbm9Ac3R5bGVyLmxpbmsiLCJlbWFpbF92ZXJpZmllZCI6Z'
        'mFsc2UsImZpcmViYXNlIjp7ImlkZW50aXRpZXMiOnsiZW1haWwiOlsiYnJ1bm8uc3V'
        'nYW5vQHN0eWxlci5saW5rIl19LCJzaWduX2luX3Byb3ZpZGVyIjoiY3VzdG9tIn19.'
        'VNJtVQ9nmvbpUkMM95cdN06O3NDUqdBwEgmczfWKR2Vt5fOFQGtFykEovEmajdVHZO'
        'm4tbJlbfoILu0i_GsO-7jaoYNBOC9-tIBiD8pvUCeL4iQj2jhpvR6z3PEtTyA7i7Ae'
        'VUj9wuxrzYm0wtcLTR1vSEp4jTj3Nw5KD4adPq4IF63SKqywNLkmgURuQnF-WlaZBn'
        'g2CK0AMddY0Qt-1Z0IFUp-i_tXkGMOPBjXswq1kPxyZ2xtNWZjQZ5sWNw2dmkC2wyf'
        'EmnQ5Zm9UJvrS6tiCigTGkF43PhwTl26K9ztwOtekrpOiwap-c55rUG8TbobpmGGkF'
        'i9Q3r5dHH-ag'
    )

    def test_get_identity(self):
        request = Mock()
        request.headers.get.return_value = f'Bearer {self.custom_token}'
        contr = BaseController()

        identity = contr.get_identity(request)

        assert isinstance(identity, RequestScope)

    def test_invalid_jwt(self):
        request = Mock()
        request.headers.get.return_value = 'Bearer aa.aa.aa'
        contr = BaseController()
        contr.bad_request = MagicMock()

        _ = contr.get_identity(request)

        contr.bad_request.assert_called_once()



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    'aiohttp',
    'sqlalchemy',
    'pyyaml',
    'psycopg2-binary',
    'sqlalchemy-serializer',
    'styler-identity',
    'styler-middleware',
    'styler-validation',
    'cachetools',
    'certifi',
    'chardet',
    'google-api-core[grpc]',
    'google-auth',
    'google-cloud-core',
    'google-cloud-monitoring',
    'google-cloud-trace',
    'googleapis-common-protos',
    'grpcio',
    'idna',
    'opencensus-context',
    'opencensus-ext-stackdriver',
    'opencensus',
    'protobuf',
    'pyasn1-modules',
    'pyasn1',
    'pytz',
    'requests',
    'rsa',
    'six',
    'urllib3',
    'google.cloud.logging',
    'python-i18n[YAML]',
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Bruno Toshio Sugano",
    author_email='brunotoshio@gmail.com',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ],
    description="Standards used in REST services",
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    include_package_data=True,
    keywords='styler_rest_framework',
    name='styler_rest_framework',
    packages=find_packages(include=['styler_rest_framework', 'styler_rest_framework.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/STYLER-Inc/styler-rest-framework',
    version='0.5.0',
    zip_safe=False,
)



""" Define helper methods to reduce the boilerplate code for a server
"""

import logging

from styler_middleware import handle_exceptions, handle_invalid_json
from styler_rest_framework.logging import setup_logging


def add_middlewares(app):
    """ Append default middlewares
    """
    app.middlewares.extend([handle_exceptions(), handle_invalid_json()])


def set_logging():  # pragma: no coverage
    setup_logging(logging.INFO)


def default_configuration(app):  # pragma: no coverage
    set_logging()
    add_middlewares(app)



""" Tests for the session aware decorator
"""

from contextlib import contextmanager
from unittest.mock import Mock, patch

from styler_rest_framework.business import session_aware
import pytest


@contextmanager
def mock_session(engine):
    yield 'db session'


@patch('styler_rest_framework.datasource.session_scope.create', mock_session)
def test_session_aware():
    @session_aware
    def mock_func(obj, db_session=None):
        return db_session
    my_business = Mock()
    my_business.db_engine = Mock()

    result = mock_func(my_business)

    assert result == 'db session'


def mock_session2(m):
    @contextmanager
    def create(engine):
        yield m
    return create


def test_close():
    @session_aware
    def mock_func(obj, db_session=None):
        raise ValueError()
    my_business = Mock()
    my_business.db_engine = Mock()
    mocked = Mock()
    with patch(
        'styler_rest_framework.datasource.session_scope.create',
        mock_session2(mocked)
    ):
        with pytest.raises(ValueError):
            _ = mock_func(my_business)

    mocked.close.assert_called_once()



""" Errors raised by business classes
"""


class BusinessError(Exception):  # pragma: no coverage
    """ Error raised from business classes
    """


class InternalError(BusinessError):  # pragma: no coverage
    """ Error raised when an unexpected error occurs
    """


class PermissionDeniedError(BusinessError):  # pragma: no coverage
    """ Error raised when the user has no access to the resource
    """


class ResourceNotFoundError(BusinessError):  # pragma: no coverage
    """ Error raised when the required resource was not found
    """


class ValidationError(BusinessError):   # pragma: no coverage
    """ Error raised when there is a validation error
    """
    def __init__(self, errors):
        super().__init__()
        self.errors = errors



""" Services module
"""

import logging

from aiohttp import ClientSession
from styler_rest_framework.exceptions.services import (
    AuthenticationError,
    AuthorizationError,
    InternalServerError,
    InvalidDataError,
    NotFoundError,
    PaymentRequiredError,
    UnexpectedError,
)


class HTTPHandler:
    """ Handles HTTP operations
    """
    def __init__(self, identity, headers=None):
        self.identity = identity
        if headers is None:
            headers = {}
        self.headers = {
            'Authorization': f'Bearer {identity.token()}',
            **{'Accept-Language': identity.localization()},
            **identity.trace_header(),
            **headers
        }

    async def post(self, url, params, error_handlers=None, **kwargs):
        headers = self._prepare_headers(kwargs.get('headers', {}))
        async with ClientSession() as session:
            async with session.post(url,
                                    headers=headers,
                                    json=params,
                                    ssl=True) as resp:
                if resp.status not in [200, 201]:
                    if error_handlers and resp.status in error_handlers:
                        await error_handlers[resp.status](resp)
                    await self._handle_http_errors(resp)

                return await resp.json()

    async def patch(self, url, params, error_handlers=None, **kwargs):
        headers = self._prepare_headers(kwargs.get('headers', {}))
        async with ClientSession() as session:
            async with session.patch(url,
                                     headers=headers,
                                     json=params,
                                     ssl=True) as resp:
                if resp.status not in [200]:
                    if error_handlers and resp.status in error_handlers:
                        await error_handlers[resp.status](resp)
                    await self._handle_http_errors(resp)

                return await resp.json()

    async def put(self, url, params, error_handlers=None, **kwargs):
        headers = self._prepare_headers(kwargs.get('headers', {}))
        async with ClientSession() as session:
            async with session.put(url,
                                   headers=headers,
                                   json=params,
                                   ssl=True) as resp:
                if resp.status not in [200]:
                    if error_handlers and resp.status in error_handlers:
                        await error_handlers[resp.status](resp)
                    await self._handle_http_errors(resp)

                return await resp.json()

    async def delete(self, url, error_handlers=None, **kwargs):
        headers = self._prepare_headers(kwargs.get('headers', {}))
        async with ClientSession() as session:
            async with session.delete(url, headers=headers, ssl=True) as resp:
                if resp.status not in [200]:
                    if error_handlers and resp.status in error_handlers:
                        await error_handlers[resp.status](resp)
                    await self._handle_http_errors(resp)

                return await resp.json()

    async def get(self, url, error_handlers=None, **kwargs):
        headers = self._prepare_headers(kwargs.get('headers', {}))
        async with ClientSession() as session:
            async with session.get(url, headers=headers, ssl=True) as resp:
                if resp.status != 200:
                    if error_handlers and resp.status in error_handlers:
                        await error_handlers[resp.status](resp)
                    await self._handle_http_errors(resp)

                return await resp.json()

    def _prepare_headers(self, overrides):
        headers = self.headers
        headers.update(overrides)
        return headers

    async def _handle_http_errors(self, resp):
        """ Handles HTTP errors
        """
        response_text = await resp.text()
        if resp.status == 400:
            raise InvalidDataError(resp.status, response_text)
        elif resp.status == 401:
            raise AuthenticationError(resp.status, response_text)
        elif resp.status == 402:
            raise PaymentRequiredError(resp.status, response_text)
        elif resp.status == 403:
            raise AuthorizationError(resp.status, response_text)
        elif resp.status == 404:
            raise NotFoundError(resp.status, response_text)
        elif resp.status == 500:
            raise InternalServerError(resp.status, response_text)
        else:
            logging.error('Unexpected response code: %s', resp.status)
            raise UnexpectedError(resp.status, response_text)
